# Copyright (C) 2009, 2010  Rickard Lindberg, Roger Lindberg
#
# This file is part of Timeline.
#
# Timeline is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Timeline is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Timeline.  If not, see <http://www.gnu.org/licenses/>.


import datetime
from datetime import time
from datetime import datetime as dt

import wx
from wx.lib.masked import TimeCtrl

from timelinelib.db.objects import TimePeriod


class DateTimePicker(wx.Panel):
    """
    Control to pick a Python datetime object.

    The time part will default to 00:00:00 if none is entered.
    """

    def __init__(self, parent, show_time=True):
        wx.Panel.__init__(self, parent)
        self.min = wx.DateTimeFromDMY(1, wx.DateTime.Jan, 10)
        self.max = wx.DateTimeFromDMY(1, wx.DateTime.Jan, 9990)
        self._create_gui()
        self.show_time(show_time)

    def show_time(self, show=True):
        self.time_picker.Show(show)
        self.GetSizer().Layout()

    def get_value(self):
        """
        Return the selected date time as a Python datetime object or raise
        ValueError if no valid date is selected.
        """
        # Move to the next field so that the text field in the
        # GenericDatePickerCtrl looses focus. The date retrieved with GetValue
        # only gets updated when the text field looses focus. If this control
        # is used in a dialog we can have problems if we don't do this. See
        # http://stackoverflow.com/questions/1568491/wx-datepickerctrl-in-dialog-ignores-value-entered-after-hitting-return-on-wxgtk
        # for more info.
        self.date_picker.Navigate()
        date = self.date_picker.GetValue()
        if not date.IsValid():
            raise ValueError(_("Invalid date."))
        if date < self.min or date > self.max:
            raise ValueError(_("Date not within allowed period."))
        date_time = dt(date.Year, date.Month+1, date.Day)
        if self.time_picker.IsShown():
            time = self.time_picker.GetValue(as_wxDateTime=True)
            date_time = date_time.replace(hour=time.Hour, minute=time.Minute)
        return date_time

    def set_value(self, value):
        if value == None:
            now = dt.now()
            value = dt(now.year, now.month, now.day)
        wx_date_time = self._python_date_to_wx_date(value)
        self.date_picker.SetValue(wx_date_time)
        self.time_picker.SetValue(wx_date_time)

    def _create_gui(self):
        # ALLOWNONE will allow us to enter invalid dates. When we do that the
        # field is filled with just blanks. The get_value method will raise
        # ValueErrors and it is up to the calling code to handle those.
        style = wx.DP_DROPDOWN | wx.DP_SHOWCENTURY | wx.DP_ALLOWNONE
        self.date_picker = wx.GenericDatePickerCtrl(self, style=style)
        self.time_picker = TimeCtrl(self, format="24HHMM")
        # Layout
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self.date_picker, proportion=1,
                  flag=wx.ALIGN_CENTER_VERTICAL)
        sizer.Add(self.time_picker, proportion=0,
                  flag=wx.ALIGN_CENTER_VERTICAL)
        self.SetSizerAndFit(sizer)

    def _python_date_to_wx_date(self, py_date):
        return wx.DateTimeFromDMY(py_date.day, py_date.month-1, py_date.year,
                                  py_date.hour, py_date.minute,
                                  py_date.second)
